/***********************************************
*
* Copyright 2011 by Sean Conner.  All Rights Reserved.
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*
* Comments, questions and criticisms can be sent to: sean@conman.org
*
************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include <assert.h>

#include <X11/Xlib.h>
#include <X11/Xos.h>
#include <X11/Xutil.h>
#include <X11/Xatom.h>

#include <cgilib/memory.h>
#include <cgilib/ddt.h>
#include <cgilib/stream.h>
#include <cgilib/errors.h>
#include <cgilib/url.h>
#include <cgilib/http.h>
#include <cgilib/htmltok.h>
#include <cgilib/util.h>

/******************************************************************/

struct ucs_ascii
{
  unsigned short  val;
  char           *text;
};

/*****************************************************************/

static int	 ucs2_ascii_cmp		(const void *,const void *);
static void	 cleanup		(void);
static char	*convert_ucs2_ascii	(unsigned short *,size_t);
static void	 burn_page		(Stream);
static char	*get_title		(const char *);

/********************************************************************/

const char *cheaders[] = 
{
  "User-agent: Blockquote/1.0 (http://boston.conman.org/)",
  "Connection: close",
  NULL
};

const struct ucs_ascii cconversion[] =
{
  {   160	, "&nbsp;"	} ,
  {   161 	, "&iexcl;"	} ,
  {   162	, "&cent;"	} ,
  {   163	, "&pound;"	} ,
  {   164	, "&curren;"	} ,
  {   165	, "&yen;"	} ,
  {   166	, "&brvbar;"	} ,
  {   167	, "&sect;"	} ,
  {   168	, "&uml;"	} ,
  {   169	, "&copy;"	} ,
  {   170	, "&ordf;"	} ,
  {   171	, "&laquo;"	} ,
  {   172	, "&not;"	} ,
  {   173	, "&shy;"	} ,
  {   174	, "&reg;"	} ,
  {   175	, "&macr;"	} ,
  {   176	, "&deg;"	} ,
  {   177	, "&plusmn;"	} ,
  {   178	, "&sup2;"	} ,
  {   179	, "&sup3;"	} ,
  {   180	, "&acute;"	} ,
  {   181	, "&micro;"	} ,
  {   182	, "&para;"	} ,
  {   183	, "&middot;"	} ,
  {   184	, "&cedil;"	} ,
  {   185	, "&sup1;"	} ,
  {   186	, "&ordm;"	} ,
  {   187	, "&raquo;"	} ,
  {   188	, "&frac14;"	} ,
  {   189	, "&frac12;"	} ,
  {   190	, "&frac34;"	} ,
  {  8211	, "--"		} ,
  {  8212	, "---"		} ,
  {  8216	, "&lsquo;"	} ,
  {  8217	, "&rsquo;"	} ,
  {  8220	, "``"		} ,
  {  8221	, "''"		} ,
  {  8230	, "..."		}
};

#define CMAX	(sizeof(cconversion) / sizeof(struct ucs_ascii))

Display *gdisplay;
Window   gwindow;
Atom     ghtml;
Atom     gurl;
Atom     gtarget;

/******************************************************************/

int main(int argc,char *argv[])
{
  unsigned long  itemcount;
  unsigned long  bytes;
  unsigned char *data;
  Atom           type;
  XEvent         event;
  int            format;
  char          *turl;
  char          *thtml;
  char          *title;
  int            rc;

  MemInit();
  DdtInit();
  StreamInit();
  
  atexit(cleanup);
  
  gdisplay = XOpenDisplay(NULL);
  if (gdisplay == NULL)
  {
    fprintf(stderr,"forgot to export DISPLAY?\n");
    return EXIT_FAILURE;
  }
  
  gwindow = XCreateSimpleWindow(
  		gdisplay,
  		RootWindow(gdisplay,DefaultScreen(gdisplay)),
  		0,
  		0,
  		10,
  		10,
  		1,
  		0,
  		0
  	);
  	
  ghtml   = XInternAtom(gdisplay,"text/html",True);
  gurl    = XInternAtom(gdisplay,"text/x-moz-url-priv",True);
  gtarget = XInternAtom(gdisplay,"XSELECT",False);
  
  if ((ghtml == None) || (gurl == None))
  {
    fprintf(stderr,"forgot to run Firefox and make a selection?\n");
    return EXIT_FAILURE;
  }
  
  XConvertSelection(
  	gdisplay,
  	XA_PRIMARY,
  	gurl,
  	gtarget,
  	gwindow,
  	CurrentTime
  );
  
  do
    XNextEvent(gdisplay,&event);
  while (event.type != SelectionNotify);
  
  if (event.xselection.property == None)
  {
    fprintf(stderr,"Can't convert\n");
    return EXIT_FAILURE;
  }
  
  rc = XGetWindowProperty(
  		event.xselection.display,
  		event.xselection.requestor,
  		event.xselection.property,
  		0,
  		LONG_MAX / 4,
  		True,
  		AnyPropertyType,
  		&type,
  		&format,
  		&itemcount,
  		&bytes,
  		&data
  	);
  
  if ((rc == 1) || (format != 8) || (type != gurl))
  {
    fprintf(stderr,"failed, wrong format or type\n");
    return EXIT_FAILURE;
  }
  
  turl = convert_ucs2_ascii((unsigned short *)data,itemcount / 2);
  title = get_title(turl);
  if (title == NULL)
    title = "(untitled)";

  XConvertSelection(
  	gdisplay,
  	XA_PRIMARY,
  	ghtml,
  	gtarget,
  	gwindow,
  	CurrentTime
  );
  
  do
    XNextEvent(gdisplay,&event);
  while (event.type != SelectionNotify);
  
  if (event.xselection.property == None)
  {
    fprintf(stderr,"can't convert\n");
    return EXIT_FAILURE;
  }
  
  rc = XGetWindowProperty(
  		event.xselection.display,
  		event.xselection.requestor,
  		event.xselection.property,
  		0,
  		LONG_MAX / 4,
  		True,
  		AnyPropertyType,
  		&type,
  		&format,
  		&itemcount,
  		&bytes,
  		&data
  	);
  
  if ((rc == 1) || (format != 8) || (type != ghtml))
  {
    fprintf(stderr,"failed, wrong format or type\n");
    return EXIT_FAILURE;
  }
  
  thtml = convert_ucs2_ascii((unsigned short *)data,itemcount / 2);
  
  assert(turl  != NULL);
  assert(title != NULL);
  assert(thtml != NULL);
  
  LineSFormat(
  	StdoutStream,
  	"$ $ $",
  	"<blockquote\n"
  	"\tcite=\"%a\"\n"
  	"\ttitle=\"%b\"\n"
  	">\n"
  	"\n"
  	"\t%c\n"
  	"\n"
  	"\t<p class=\"cite\"><cite><a class=\"external\" href=\"%a\">"
  		"%b</a></cite></p>\n"
  	"\n"
  	"</blockquote>\n"
  	"\n",
  	turl,
  	title,
  	thtml
  );
  
  return EXIT_SUCCESS;
}

/*******************************************************************/

static void cleanup(void)
{
  if (gwindow)  XDestroyWindow(gdisplay,gwindow);
  if (gdisplay) XCloseDisplay(gdisplay);
}

/*********************************************************************/

static int ucs2_ascii_cmp(const void *key,const void *target)
{
  const unsigned short   *k = key;
  const struct ucs_ascii *t = target;
  
  assert(key    != NULL);
  assert(target != NULL);

  if (*k < t->val)
    return -1;
  else if (*k > t->val)
    return  1;
  else
    return  0;
}

/**********************************************************************/

static char *convert_ucs2_ascii(unsigned short *str,size_t len)
{
  Stream            s;
  struct ucs_ascii *conv;
  char             *text;
  
  assert(str != NULL);
  assert(len >  0);
  
  s = StringStreamWrite();
  
  while(len--)
  {
    if (*str == 0xFEFF)
      ;
    else if (*str == 0xFFFE)
      ;
    else if (*str == '\n')
      StreamWrite(s,' ');
    else if (*str <= 126)
      StreamWrite(s,*str);
    else
    {
      conv = bsearch(str,cconversion,CMAX,sizeof(struct ucs_ascii),ucs2_ascii_cmp);
      if (conv == NULL)
        LineS(s,"???");
      else
        LineS(s,conv->text);
    }
    str++;
  }
  
  text = StringFromStream(s);
  StreamFree(s);
  return text;
}

/********************************************************************/

static void burn_page(Stream in)
{
  Stream null;
  
  assert(in != NULL);
  
  null = StreamNewWrite();
  StreamCopy(null,in);
  StreamFree(null);
}

/*****************************************************************/
  
static char *get_title(const char *turl)
{
  URLHTTP    http;
  HTTP       conn;
  HtmlToken  token;
  int        t;
  char      *title;
  
  if (UrlNew((URL *)&http,turl) != ERR_OKAY) return NULL;
  if (HttpConnect(&conn,GET,http,cheaders) != ERR_OKAY) 
  {
    UrlFree((URL *)&http);
    return NULL;
  }
  if (HttpStatus(conn) != 200)
  {
    /*------------------------------------------------
    ; let's at least be nice and get the page before
    ; discarding it ... 
    ;-----------------------------------------------*/
    
    burn_page(HttpStreamRead(conn));
    HttpClose(&conn);
    UrlFree((URL *)&http);
    return NULL;
  }
  
  if (HtmlParseNew(&token,HttpStreamRead(conn)) != ERR_OKAY)
  {
    burn_page(HttpStreamRead(conn));
    HttpClose(&conn);
    UrlFree((URL *)&http);
    return NULL;
  }
  
  while((t = HtmlParseNext(token)) != T_EOF)
  {
    if ((t == T_TAG) && (strcmp(HtmlParseValue(token),"TITLE") == 0))
    {
      t = HtmlParseNext(token);
      if (t != T_STRING)
        title = NULL;
      else
        title = dup_string(HtmlParseValue(token));
      
      break;
    }
  }
  
  
  HtmlParseFree(&token);
  burn_page(HttpStreamRead(conn));
  HttpClose(&conn);
  UrlFree((URL *)&http);
  return title;
}

/********************************************************************/

